#include "communication.h"
#include "commands.h"
#include "type_def.h"
#include "uart.h"

/*
 * Static variables used to keep track of packet counts
 */
static int n_msg_received = 0;
static size_t total_payload_received = 0;


/* Misc. callbacks */

/**
  * Currently does nothing.
  */
int debug(modular_structs_t *structs)
{
	return 0;
}

/**
  * counts the number of packet logs.
  */
int cb_packetlog(modular_structs_t* structs) {
	n_msg_received += 1;
	total_payload_received += uart_buff_data_length();
	return 0;
}

/**
  * Handles a get packet logs request and sends a response
  * with the packet log data.
  */
int cb_getpacketlogs(modular_structs_t* structs) {
	char buf[255];

	// Message logging number of messages received and size of payload received
	int length = snprintf(buf, sizeof buf, "%d,%d", n_msg_received, total_payload_received);

	send_data(LOG_ID, 0, buf, length >= sizeof(buf) ? 255 : length + 1);
	return 0;
}

/*
 * Handles receiving new location updates.
 */
int cb_update(modular_structs_t *structs)
{
	//processUpdate(packet, &(structs->raw_sensor_struct.currentQuadPosition));

	quadPosition_t* currentQuadPosition = &(structs->raw_sensor_struct.currentQuadPosition);
	// Packet must come as [NEARPY], 4 bytes each
	int packetId = uart_buff_data_get_u32(0);
//	printf("Packet ID: %d\n", packetId);
	float y_pos = uart_buff_data_get_float(4);
//	printf("y_pos: %f\n", y_pos);
	float x_pos = uart_buff_data_get_float(8);
//	printf("x_pos: %f\n", x_pos);
	float alt_pos = uart_buff_data_get_float(12);
//	printf("alt_pos: %f\n", alt_pos);
	float roll = uart_buff_data_get_float(16);
//	printf("roll: %f\n", roll);
	float pitch = uart_buff_data_get_float(20);
//	printf("pitch: %f\n", pitch);
	float yaw = uart_buff_data_get_float(24);
//	printf("yaw: %f\n", yaw);

	currentQuadPosition->packetId = packetId;
	currentQuadPosition->y_pos = y_pos;
	currentQuadPosition->x_pos = x_pos;
	currentQuadPosition->alt_pos = alt_pos;
	currentQuadPosition->roll = roll;
	currentQuadPosition->pitch = pitch;
	currentQuadPosition->yaw = yaw;

	// Make location as fresh
	structs->user_input_struct.locationFresh = 1;

	return 0;
}

/**
  * This is called on the ground station to begin sending VRPN to the quad.
  */
int cb_beginupdate(modular_structs_t *structs) {
	structs->user_input_struct.receivedBeginUpdate = 1;
	return 0;
}


/* Callbacks for configuration */

/**
  * Handles a command to set a controller parameter on the quad.
  *
  * NOTE:
  * Expects the uart buff to have data in the following format:
  * |--------------------------------------------------------|
  * |  data index ||      0      |      1      |    2 - 5    |
  * |--------------------------------------------------------|
  * |   parameter ||  control ID | ctrl parmID |  param val  |
  * |--------------------------------------------------------|
  * |       bytes ||      1      |      1      |      4      |
  * |--------------------------------------------------------|
  * 
  * Does not send anything in response.
  */
int cb_setparam(modular_structs_t *structs)
{
	// Get some of the meta data
	u16 data_len = uart_buff_data_length();
	// Check if the data length is correct
	if (data_len != 6)
	{
		return -1;
	}

	// Get the controller ID, parameter ID, parameter value
	u8 controller_id = uart_buff_data_get_u8(0);
	u8 param_id = uart_buff_data_get_u8(1);
	float param_val = uart_buff_data_get_float(2);
	// Check to make sure the IDs are in bounds
	if (controller_id >= MAX_CONTROLLER_ID ||
		param_id >= MAX_CONTROL_PARAM_ID)
	{
		return -1;
	}

	// Set the param_val into the controller by controller_id, param_id
	switch(param_id)
	{
		case KP_ID:
			structs->parameter_struct.pid_controllers[controller_id].Kp = param_val;
			break;
		case KI_ID:
			structs->parameter_struct.pid_controllers[controller_id].Ki = param_val;
			break;
		case KD_ID:
			structs->parameter_struct.pid_controllers[controller_id].Kd = param_val;
			break;
		case SP_ID:
			structs->parameter_struct.pid_controllers[controller_id].setpoint = param_val;
			break;
	}

	return 0;
}

/**
  * Handles a command to get a controller parameter from the quad.
  *
  * NOTE:
  * Expects the uart buff to have data in the following format:
  * |------------------------------------------|
  * |  data index ||      0      |      1      |
  * |------------------------------------------|
  * |   parameter ||  control ID | ctrl parmID |
  * |------------------------------------------|
  * |       bytes ||      1      |      1      |
  * |------------------------------------------|
  *
  * Sends a response of type RESPONSECONTROL_ID.
  * The response will have a message ID equal to the one originally received.
  * The data of the response will be in the following format:
  * |--------------------------------------------------------|
  * |  data index ||      0      |      1      |    2 - 5    |
  * |--------------------------------------------------------|
  * |   parameter ||  control ID | ctrl parmID |  param val  |
  * |--------------------------------------------------------|
  * |       bytes ||      1      |      1      |      4      |
  * |--------------------------------------------------------|
  */
int cb_getparam(modular_structs_t* structs)
{
	// Get some of the meta data
	u16 data_len = uart_buff_data_length();
	u16 msg_id = uart_buff_get_u16(3);
	// Check if the data length is correct
	if (data_len != 2)
	{
		return -1;
	}

	// Get the controller ID, parameter ID
	u8 controller_id = uart_buff_data_get_u8(0);
	u8 param_id = uart_buff_data_get_u8(1);
	// Check to make sure the IDs are in bounds
	if (controller_id >= MAX_CONTROLLER_ID ||
		param_id >= MAX_CONTROL_PARAM_ID)
	{
		return -1;
	}

	// Make the variable to send
	float param_val;
	// Set the param_val equal to the parameter value stored in the controller by
	// controller_id, param_id
	switch(param_id)
	{
		case KP_ID:
			param_val = structs->parameter_struct.pid_controllers[controller_id].Kp;
			break;
		case KI_ID:
			param_val = structs->parameter_struct.pid_controllers[controller_id].Ki;
			break;
		case KD_ID:
			param_val = structs->parameter_struct.pid_controllers[controller_id].Kd;
			break;
		case SP_ID:
			param_val = structs->parameter_struct.pid_controllers[controller_id].setpoint;
			break;
	}

	// Format the response data
	char resp_data[6];
	// Controller ID
	resp_data[0] = controller_id;
	// Parameter ID
	resp_data[1] = param_id;
	// Parameter value (4 byte float)
	// TODO set a strict byte ordering for communication between the ground station and the quad
	memcpy(&resp_data[2], &param_val, sizeof(param_val));

	// Send the response
	send_data(RESPPARAM_ID, msg_id, resp_data, sizeof(resp_data));

	return 0;
}
